#include <stdio.h>
#include <stdlib.h>

#include <pthread.h>

int n, num_threads;
double d, pi;

pthread_mutex_t reduction_mutex;
pthread_t *tid;

void *PIworker(void *arg)
{
  int i, myid;
  double s, x, mypi;
  
  myid = *(int *)arg;
  s = 0.0;
  
  for (i=myid+1; i<=n; i+=num_threads) {
    x = (i-0.5)*d;
    s += 4.0/(1.0+x*x);
  }

  mypi = d*s;
  
  pthread_mutex_lock(&reduction_mutex);
  pi += mypi;
  pthread_mutex_unlock(&reduction_mutex);
  
  pthread_exit(0);
}

int main (int argc, char **argv)
{
  int i;
  int *id;

  if (argc != 3) {
    n = 10;
    num_threads = 1;
  }
  else {
    n = atoi (argv[1]);
    num_threads = atoi(argv[2]);
  }

  d = 1.0 / n;
  pi = 0.0;

  if ( (id = (int *) calloc(n,sizeof(int))) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    exit (EXIT_FAILURE);
  }

  if ( (tid = (pthread_t *) calloc(num_threads, sizeof(pthread_t))) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    exit (EXIT_FAILURE);
  }

  if(pthread_mutex_init (&reduction_mutex, NULL)) {
    fprintf (stderr, "Cannot init lock\n");
    exit (EXIT_FAILURE); 
  };

  for (i=0; i<num_threads; i++) {
    id[i] = i;
    if(pthread_create(&tid[i],NULL,
                          PIworker,(void *)&id[i])) {
      exit (EXIT_FAILURE); 
    } 
  }

  for (i=0; i<num_threads; i++)
    pthread_join(tid[i],NULL);
  pthread_mutex_destroy (&reduction_mutex);

  fprintf (stdout, "pi = %.15f\n", pi);

  free (id);
  free (tid);

  return EXIT_SUCCESS;
}
